from cubes.container import CONTAINERS


wf_rtype_roles = (('custom_workflow', 'subject'),
                  ('in_state', 'subject'),
                  ('wf_info_for', 'object'))

for etype, cfg in CONTAINERS.iteritems():
    if hasattr(cfg, 'collaboration_config'):
        eschema = schema[etype]
        if not any(eschema.has_relation(rtype, role)
                   for rtype, role in wf_rtype_roles):
            make_workflowable(etype)
        default = True
        if rql('Any X,W WHERE X name "%s", X default_workflow W' % etype):
            print ('%s already has a default worklow, collaboration '
                   'workflow will not be the default one.' % etype)
            default = False
        collabcfg = cfg.collaboration_config
        wf = collabcfg.build_collaboration_workflow(session, default=default)
        for entity in find_entities(etype):
            adapted = entity.cw_adapt_to('IWorkflowable')
            state = u'wfs_immutable' if getattr(entity, 'frozen', False) else u'wfs_mutable'
            adapted.set_initial_state(state)
        drop_attribute(etype, 'frozen')



# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-collaboration automatic tests


uncomment code below if you want to activate automatic test for your cube:

.. sourcecode:: python

    from cubicweb.devtools.testlib import AutomaticWebTest

    class AutomaticWebTest(AutomaticWebTest):
        '''provides `to_test_etypes` and/or `list_startup_views` implementation
        to limit test scope
        '''

        def to_test_etypes(self):
            '''only test views for entities of the returned types'''
            return set(('My', 'Cube', 'Entity', 'Types'))

        def list_startup_views(self):
            '''only test startup views of the returned identifiers'''
            return ('some', 'startup', 'views')
"""

from cubicweb.devtools import testlib

class DefaultTC(testlib.CubicWebTC):
    def test_something(self):
        self.skipTest('this cube has no test')


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



"""cubicweb-collaboration application package

Cube for providing facilities for cooperative entity handling
"""
from warnings import warn
from itertools import chain
import logging

from rql.utils import is_keyword
from yams.schema import PermissionMixIn, EntitySchema, RelationDefinitionSchema
from yams.buildobjs import RelationDefinition, RelationType, DEFAULT_RELPERMS

# Until https://www.logilab.org/ticket/245548
from yams.buildobjs import EntityType
DEFAULT_ETYPEPERMS = EntityType.__permissions__

from cubicweb.predicates import relation_possible, is_in_state
from cubicweb.schema import (RQLConstraint, ERQLExpression, RRQLExpression,
                             WORKFLOW_RTYPES, WORKFLOW_TYPES,
                             split_expression)

from cubes.container import ContainerConfiguration


LOGGER = logging.getLogger('cubes.collaboration')


# Additionnal methods to yams.schema classes (EntitySchema and
# RelationDefinitionSchema).

# rql.utils.KEYWORDS misses a comma, fix it here temporarily.
# see https://www.logilab.org/ticket/245563
from rql.utils import KEYWORDS
if 'NOTIN' in KEYWORDS:
    KEYWORDS.remove('NOTIN')
    KEYWORDS.update(['NOT', 'IN'])


def get_rqlexpr_vars(expression, mainvars=()):
    return set(w.rstrip('?') for w in split_expression(expression)
               if w.isupper() and not is_keyword(w) and w not in mainvars)


def _add_mandatory_rqlexpression(perms, expression, mainvars, expr_cls):
    """Combine `expression` with existing RQL expressions in `perms`"""
    has_rqlexpr = False
    for idx, perm in enumerate(perms):
        if isinstance(perm, expr_cls):
            has_rqlexpr = True
            defined = get_rqlexpr_vars(expression, mainvars=mainvars)
            expr = perm.expression
            for var in get_rqlexpr_vars(expr, mainvars=mainvars):
                if var in defined:
                    # Maybe just issue a warning there.
                    raise Exception('Variable `%(v)s` already used in RQL '
                                    'expression `%(expr)s`' %
                                    {'v': var, 'expr': expr})
                defined.add(var)
            perms[idx] = expr_cls(', '.join([expression, expr]))
    if not has_rqlexpr:
        perms.append(expr_cls(expression))
    return perms


def schema_add_mandatory_erqlexpression(self, action, expression):
    """Overload permissions of `action` with `expression` ERQLExpression."""
    perms = list(self.permissions[action])
    perms = _add_mandatory_rqlexpression(
        perms, expression, mainvars=('X', 'U'), expr_cls=ERQLExpression)
    self.set_action_permissions(action, tuple(perms))
EntitySchema.permissions_add_mandatory_rqlexpression = schema_add_mandatory_erqlexpression


def schema_add_mandatory_rrqlexpression(self, action, expression):
    """Overload permissions of `action` with `expression` RRQLExpression."""
    perms = list(self.permissions[action])
    perms = _add_mandatory_rqlexpression(
        perms, expression, mainvars=('S', 'O', 'U'), expr_cls=RRQLExpression)
    self.set_action_permissions(action, tuple(perms))
RelationDefinitionSchema.permissions_add_mandatory_rqlexpression = schema_add_mandatory_rrqlexpression


def schema_set_permissions_groups(self, action, groups):
    """Set `groups` for `action` permissions. """
    rqlexprs = tuple(p for p in self.permissions[action]
                     if not isinstance(p, basestring))
    self.set_action_permissions(action, tuple(groups + rqlexprs))
PermissionMixIn.set_permissions_groups = schema_set_permissions_groups


# Collaboration configuration.

class CollaborationConfiguration(object):
    """Configuration object to make an entity type collaborative.

    Main methods are ``setup_schema``, ``build_hooks``,
    ``build_adapters``, ``setup_ui`` to be respectively in
    post-build/registration callbacks of schema, hooks,
    entities, views respectively.

    Notes about migration:

        *   the method ``rtypes_to_add`` returns collaboration-specific
            relation types to be added.
        *   the method ``etypes_to_sync`` returns entity types to synchronize
            because collaboration modified them (``sync_schema_props_perms``).
    """

    def __init__(self, container_config,
                 clone_rtype=None, clone_skiprtypes=(), clone_skipetypes=(),
                 clone_compulsory_hooks_categories=(), clone_state='wfs_immutable',
                 **kwargs):
        if not isinstance(container_config, ContainerConfiguration):
            warn('[0.3] CollaborationConfiguration now requires a '
                 'ContainerConfiguration as first argument',
                 DeprecationWarning, stacklevel=2)
            etype = container_config
            if 'rtype' not in kwargs:
                kwargs['rtype'] = 'in_' + etype.lower()
            container_config = ContainerConfiguration(etype, **kwargs)
        elif kwargs:
            # kwargs exists only for bw compatibility.
            raise TypeError("got unexpected keyword arguments '%s'" %
                            ', '.join(kwargs))
        self.container_config = container_config
        container_config.collaboration_config = self
        # XXX Strange to modify the container configuration in user's back...
        self.container_config.skipetypes |= WORKFLOW_TYPES
        self.container_config.skiprtypes |= WORKFLOW_RTYPES
        # For convenience.
        self.etype = self.container_config.etype
        self.rtype = self.container_config.rtype
        if clone_rtype is None:
            clone_rtype = 'clone_of_' + etype.lower()
        self.clone_rtype = clone_rtype
        self.clone_state = clone_state
        # The sets rtypes/etypes to skip for cloning include respective sets
        # for the container configuration.
        self.clone_skiprtypes = frozenset(
            clone_skiprtypes + tuple(self.container_config.skiprtypes))
        self.clone_skipetypes = frozenset(
            clone_skipetypes + tuple(self.container_config.skipetypes))
        self.clone_compulsory_hooks_categories = clone_compulsory_hooks_categories

    def setup_schema(self, schema,
                     container_rtype_permissions=None):
        """ Main setup helper. Inline a modified version of this to customize
        collaboration details to your needs.
        """
        # Setup collaboration container.
        self.container_config.define_container(schema,
                            rtype_permissions=container_rtype_permissions)
        # Ensure collaborative entity is worklowable.
        self._make_workflowable(schema, self.etype)
        # Setup all collaboration attributes and relations.
        self.setup_clone_rdef(schema)
        if schema.has_relation(self.clone_rtype):
            # Cannot remove a collaborator who has a clone.
            delperm = RRQLExpression(
                'O owned_by U, NOT EXISTS(C %s O, C owned_by S)' %
                self.clone_rtype)
        else:
            delperm = RRQLExpression('O owned_by U')
        perms = {'read': ('managers', 'users'),
                 'delete': ('managers', delperm),
                 'add': ('managers', RRQLExpression('O owned_by U'))}
        self.setup_collaborates_on(schema, perms=perms)

    @staticmethod
    def _make_workflowable(schema, etype):
        """Make `etype` workflowable if it is not.
        (Inspired by `cubicweb.schema.make_workflowable`.)
        """
        eschema = schema[etype]
        for rtype, role, tetype, card in [
                ('custom_workflow', 'subject', 'Workflow', '?*'),
                ('in_state', 'subject', 'State', '1*'),
                ('wf_info_for', 'object', 'TrInfo', '1*')]:
            if not eschema.has_relation(rtype, role):
                subj, obj = (etype, tetype) if role == 'subject' else (tetype, etype)
                rdef = RelationDefinition(subj, rtype, obj, cardinality=card)
                schema.add_relation_def(rdef)
                warn('[collaboration] adding %s to make %s entity type made '
                     'workflowable; consider making the declaration explicit '
                     'in respective schema.py' % (rdef, etype),
                     Warning, stacklevel=2)

    def rtypes_to_add(self):
        """Return the relation types to be added in migration."""
        return [self.rtype, self.clone_rtype, 'collaborates_on']

    def setup_clone_rdef(self, schema):
        if not schema.has_relation(self.clone_rtype):
            rtype = RelationType(self.clone_rtype)
            schema.add_relation_type(rtype)
        cstr = RQLConstraint('O in_state ST, ST name "%s"' % self.clone_state)
        rdef = RelationDefinition(self.etype, self.clone_rtype, self.etype,
                                  cardinality='?*', constraints=[cstr])
        schema.add_relation_def(rdef)

    def build_collaboration_workflow(self, session, default=True):
        """Build a workflow for the collaborative entity.

        The Workflow will have two states `wfs_mutable` and `wfs_immutable` and two
        transitions `wft_make_immutable` and `wft_make_mutable`.

        Set `default=True` will make this new workflow the default for the
        collaborative entity type.
        """
        wfname = u'collaboration workflow for ' + self.etype
        rset = session.execute('Any W WHERE W name "%(wfname)s",'
                               '            X name "%(etype)s",'
                               '            W workflow_of X',
                               {'wfname': wfname, 'etype': self.etype})
        if rset:
            wf = rset.get_entity(0, 0)
            warn('[collaboration] %s already has a collaboration workflow (%s)'
                 % (self.etype, wf),
                 Warning, stacklevel=2)
            return wf
        rset = session.execute('Any X WHERE X is CWEType, X name %(et)s',
                               {'et': self.etype})
        assert rset, 'cannot find an entity type named %s' % self.etype
        eid = rset[0][0]
        wf = session.create_entity(
            'Workflow', name=wfname,
            workflow_of=eid, reverse_default_workflow=eid if default else None)
        wfs_mutable = wf.add_state(_('wfs_mutable'), initial=True)
        wfs_immutable = wf.add_state(_('wfs_immutable'))
        wf.add_transition(_('wft_make_immutable'), (wfs_mutable, ), tostate=wfs_immutable,
                          requiredgroups=('managers', 'owners'))
        wf.add_transition(_('wft_make_mutable'), (wfs_immutable, ), tostate=wfs_mutable,
                          requiredgroups=('managers', ),
                          conditions=(
                              {'expr': u'X owned_by U, NOT EXISTS(C %s X)' %
                               self.clone_rtype,
                               'mainvars': u'X,U'}, ),
                         )
        return wf

    def setup_collaborates_on(self, schema, perms):
        rdef = RelationDefinition('CWUser', 'collaborates_on', self.etype,
                                  cardinality='**', __permissions__=perms.copy())
        schema.add_relation_def(rdef)


    def setup_can_read(self, schema, perms):
        rdef = RelationDefinition('CWUser', 'can_read', self.etype,
                                  cardinality='**', __permissions__=perms.copy())
        schema.add_relation_def(rdef)


    def setup_can_write(self, schema, perms):
        rdef = RelationDefinition('CWUser', 'can_write', self.etype,
                                  cardinality='**', __permissions__=perms.copy())
        schema.add_relation_def(rdef)


    def _wfperms(self, mainvar):
        """Permissions snippet concerning workflow and the clonable state."""
        return ('NOT EXISTS(%(x)s in_state ST, ST name "%(st)s")' %
                {'x': mainvar, 'st': self.clone_state})

    def setup_security(self, schema, ignore_rdefs=(), extend_perms=False):
        """Main security setup helper.

        :param schema: schema instance
        :param ignore_rdefs: sequence of relation definitions for which not to
        apply specified container_rperms / near_container_rperms.
        :param extend_perms: if True, pre-existing permissions on entity types
        and relation definitions in the scope of the collaboration container
        will be extended to include collaboration specific permissions instead
        of being overridden.
        """
        # can_read/can_write relations.
        perms = {'read': ('managers', 'users'),
                 'delete': ('managers',),
                 'add': ('managers',)}
        self.setup_can_read(schema, perms=perms)
        self.setup_can_write(schema, perms=perms)
        self.setup_container_etype_security(schema, extend_perms=extend_perms)
        self.setup_container_rtypes_security(schema, ignore_rdefs=ignore_rdefs,
                                             extend_perms=extend_perms)

    def setup_container_etype_security(self, schema, extend_perms=False):
        # Permissions for the collaborative entity (container).
        eschema = schema[self.etype]
        read_perm = 'U can_read X'
        update_perm = 'U can_write X, ' + self._wfperms('X')
        if extend_perms:
            eschema.set_permissions_groups('read', ('managers',))
            eschema.permissions_add_mandatory_rqlexpression('read', read_perm)
            for action in ('update', 'delete'):
                eschema.set_permissions_groups(action, ('managers',))
                eschema.permissions_add_mandatory_rqlexpression(
                    action, update_perm)
        else:
            if eschema.permissions != DEFAULT_ETYPEPERMS:
                warn('overriding permissions of %s entity type for '
                     'collaboration' % eschema.type)
            eschema.permissions = {
                'add': ('managers', 'users'),
                'read': ('managers', ERQLExpression(read_perm)),
                'update': ('managers', ERQLExpression(update_perm)),
                'delete': ('managers', ERQLExpression(update_perm))
            }
        # Permissions for entity types in the container.
        for etype in self.container_config.structure(schema, strict=True)[1]:
            eschema = schema[etype]
            if extend_perms:
                eschema.set_permissions_groups('read', ('managers',))
                eschema.permissions_add_mandatory_rqlexpression(
                    'read', self._etype_read_perm)
                for action in ('update', 'delete'):
                    eschema.set_permissions_groups(action, ('managers',))
                    eschema.permissions_add_mandatory_rqlexpression(
                        action, self._etype_update_perm)
            else:
                if eschema.permissions != DEFAULT_ETYPEPERMS:
                    warn('overriding permissions of %s entity type for '
                         'collaboration' % eschema.type)
                schema[etype].permissions = self.etype_permissions()
        orphans = self._orphan_etypes(schema)
        if orphans:
            LOGGER.warning('no collaboration permissions set on entity types '
                           '(%s) which may exist without a container' %
                           ', '.join(orphans))

    def _orphan_etypes(self, schema):
        """Return entity types which are not strictly in the container (i.e.
        which entities can live without being related to a container).
        """
        return (self.container_config.structure(schema)[1] -
                self.container_config.structure(schema, strict=True)[1])

    @property
    def _etype_read_perm(self):
        return 'U can_read C, X %s C' % self.rtype

    @property
    def _etype_update_perm(self):
        return ('U can_write C, X %s C, %s' %
                (self.rtype, self._wfperms('C')))

    def etype_permissions(self):
        """Return the permissions dict of entity types in the container"""
        rqlexprs ={'read': self._etype_read_perm,
                   'update': self._etype_update_perm}
        return {'add': ('managers', 'users'),
                'read': ('managers', ERQLExpression(rqlexprs['read'])),
                'update': ('managers', ERQLExpression(rqlexprs['update'])),
                'delete': ('managers', ERQLExpression(rqlexprs['update']))}

    def setup_container_rtypes_security(self, schema, ignore_rdefs=(),
                                        extend_perms=False):
        """Setup permissions on relations."""
        orphans = self._orphan_etypes(schema)
        # Structural relations pointing directly to container root.
        for rschema, role in self.container_config.structural_relations_to_container(schema):
            var = role[0].upper()
            update_perm = ', '.join(['U can_write ' + var, self._wfperms(var)])
            for rdef in rschema.rdefs.itervalues():
                if rdef not in ignore_rdefs:
                    self._set_rdef_permissions(rdef, role=role,
                                               update_perm=update_perm,
                                               extend_perms=extend_perms)
        # Other structural relations.
        for rschema, role in chain(
                self.container_config.structural_relations_to_parent(schema),
                self.container_config.border_relations(schema)):
            var = role[0].upper()
            for rdef in rschema.rdefs.itervalues():
                if getattr(rdef, role).type in orphans:
                    LOGGER.warning(
                        'no collaboration permissions set on %s as its %s '
                        'may exist without its container' % (rdef, role))
                    continue
                if rdef not in ignore_rdefs:
                    self._set_rdef_permissions(rdef, role=role,
                                               extend_perms=extend_perms)
        # Non-structural relations.
        for rschema in self.container_config.inner_relations(schema):
            for rdef in rschema.rdefs.itervalues():
                if rdef.subject.type in orphans:
                    LOGGER.warning(
                        'no collaboration permissions set on %s as its subject '
                        'may exist without its container' % rdef)
                    continue
                if rdef not in ignore_rdefs:
                    self._set_rdef_permissions(rdef, extend_perms=extend_perms)

    def _set_rdef_permissions(self, rdef, role='subject', update_perm=None,
                              extend_perms=False):
        update_perm = update_perm or self._rdef_update_perm(role[0].upper())
        if extend_perms:
            for action in ('add', 'delete'):
                rdef.set_permissions_groups(action, ('managers',))
                rdef.permissions_add_mandatory_rqlexpression(
                    action, update_perm)
        else:
            if rdef.permissions != DEFAULT_RELPERMS:
                warn('overriding permissions of `%s` for collaboration' % rdef)
            rdef.permissions = self.rdef_permissions(role)

    def _rql_to_container(self, var):
        return '%s %s C' % (var, self.container_config.rtype)

    def _rdef_update_perm(self, var):
        return ', '.join([self._rql_to_container(var),
                          'U can_write C',
                          self._wfperms('C')])

    def rdef_permissions(self, parent_role):
        """Return the permissions dict of a relation definition in the
        container. `parent_role` is the role of the parent to container
        in the relation.
        """
        var = parent_role[0].upper()
        rqlexpr = self._rdef_update_perm(var)
        return {'read':   ('managers', 'users'),
                'add':    ('managers', RRQLExpression(rqlexpr)),
                'delete': ('managers', RRQLExpression(rqlexpr))}

    def build_hooks(self, schema):
        """Return configured container hooks for collaboration."""
        self.setup_on_commit_rtypes_permissions(schema)
        container_hooks = self.container_config.build_container_hooks(schema)
        return container_hooks + (self.build_clone_hook(), )

    def build_clone_hook(self):
        """Build the clone container hook."""
        from cubes.container.hooks import CloneContainer
        from cubicweb.server.hook import Hook, match_rtype
        return type(
            self.etype + 'CloneContainer', (CloneContainer, ),
            {'__select__': Hook.__select__ & match_rtype(self.clone_rtype)})

    def build_adapters(self, schema):
        """Return configured container adapters for collaboration."""
        from cubes.collaboration.entities import (ClonableAdapter,
                                                  CollaborativeITreeAdapter)
        container_protocol = self.container_config.build_container_protocol(schema)
        clone_select = relation_possible(self.clone_rtype,
                                         target_etype=self.etype)
        clone_adapter = type(self.etype + 'CloneAdapter', (ClonableAdapter, ),
                             {'__select__': clone_select,
                              'rtypes_to_skip': self.clone_skiprtypes,
                              'etypes_to_skip': self.clone_skipetypes,
                              'clone_rtype_role': (self.clone_rtype, 'subject'),
                              'compulsory_hooks_categories':
                                self.clone_compulsory_hooks_categories,
                             })
        itree_adapter = type(self.etype + 'ITreeAdapter', (CollaborativeITreeAdapter, ),
                             {'__select__': clone_select,
                              'tree_relation': self.clone_rtype})
        return container_protocol, clone_adapter, itree_adapter

    def setup_on_commit_rtypes_permissions(self, schema):
        """Extend security setup: need to check relations using relation to
        container in their perms on commit.
        """
        from cubicweb.server import ON_COMMIT_ADD_RELATIONS
        for rschema, _ in chain(
                self.container_config.structural_relations_to_container(schema),
                self.container_config.structural_relations_to_parent(schema),
                self.container_config.border_relations(schema)):
            ON_COMMIT_ADD_RELATIONS.add(rschema.type)
        for rschema in self.container_config.inner_relations(schema):
            ON_COMMIT_ADD_RELATIONS.add(rschema.type)

    def setup_ui(self):
        from cubicweb.web.views import uicfg
        _pvs = uicfg.primaryview_section
        _pvs.tag_subject_of(('*', self.rtype, '*'), 'hidden')
        _pvs.tag_object_of(('*', self.rtype, '*'), 'hidden')
        _afs = uicfg.autoform_section
        _afs.tag_subject_of(('*', self.rtype, self.etype), 'main', 'hidden')
        _afs.tag_object_of(('*', self.rtype, self.etype), 'main', 'hidden')
        _afs.tag_subject_of((self.etype, self.clone_rtype, self.etype),
                            'main', 'hidden')
        _afs.tag_object_of((self.etype, self.clone_rtype, self.etype),
                           'main', 'hidden')



# -*- coding: utf-8 -*-
# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

from contextlib import contextmanager


class CollaborationMixin(object):

    @contextmanager
    def userlogin(self, *args):
        cnx = self.login(*args)
        yield cnx
        self.restore_connection()

    @contextmanager
    def do_as_user(self, *args):
        '''
        Context manager that connects as user - given in positional
        arguments (see ``CubicWebTC.login``). The request object is
        returned. A commit is performed at exit.

        :Note: for convenience, if ``*args`` is a unique eid, the
               login is automatically retrieved.
        '''
        if len(args) == 1 and isinstance(args[0], int):
            login = self.request().entity_from_eid(args[0]).login
            args = (login,)

        cnx = self.login(*args)
        yield cnx.request()
        cnx.commit()
        self.restore_connection()

    def prepare_entities(self, etype, subetypes=(), otheretypes=()):
        req = self.request()
        self.create_user(req, 'author', groups=('users',), commit=True)
        for i in xrange(3):
            self.create_user(req, 'my_uname_%s' % i, groups=('users',), commit=True)
        with self.do_as_user('author') as req:
            self.ce_eid = req.create_entity(etype, name=u'My test entity',
                                            description=u'My description').eid
            for i in xrange(3):
                for subetype in subetypes:
                    child = req.create_entity(
                        subetype,
                        name=u'child %s %s' % (subetype, i),
                        description=u'Some description %s' % i,
                        reverse_child_entities=self.ce_eid
                    )
                    child_eid = child.eid
                    self.commit()
                    for j in xrange(3):
                        req.create_entity(
                            'Sub' + subetype,
                            name=u'sub-child %s %s of %s' % (subetype, j, i),
                            description=u'description %s of %s' % (j, i),
                            reverse_sub_children=child_eid
                        )
                for otheretype in otheretypes:
                    req.create_entity(
                        otheretype,
                        name=u'other %s %s' % (otheretype, i),
                        other_description=u'other description %s' % i,
                        reverse_other_entities=self.ce_eid
                    )
                    for j in xrange(3):
                        req.create_entity(
                            otheretype,
                            name=u'sub-other %s %s of %s' % (otheretype, j, i),
                            other_description=u'description %s of %s' % (j, i),
                            reverse_others=child_eid
                        )

    def users(self, rex=None):
        if rex is None:
            pattern = 'LIKE "my_uname_%"'
        elif rex == 'author':
            pattern = '"author"'
        else:
            pattern = 'REGEXP "my_uname_%s"' % rex
        rql = ('Any X WHERE X is CWUser, X login %(pattern)s' %
               {'pattern': pattern})
        return [row[0] for row in self.request().execute(rql)]

    def _setup_ce(self, req, collaborators=None,
                  **extra_attributes):
        entity = req.entity_from_eid(self.ce_eid)
        if extra_attributes:
            entity.cw_set(**extra_attributes)
        adapted = entity.cw_adapt_to('ICollaborative')
        if collaborators is not None:
            adapted.set_collaborators(collaborators)

    def setup_ce(self, as_user='author', collaborators=None, **extra_attributes):
        with self.do_as_user(as_user) as req:
            self._setup_ce(req,
                           collaborators=collaborators,
                           **extra_attributes)

    def assert_setup_ce_raises(self, exception, as_user=None, collaborators=None,
                               **extra_attributes):
        if as_user is not None:
            login = self.request().entity_from_eid(as_user).login
        else:
            login = 'author'
        with self.userlogin(login) as cnx:
            req = cnx.request()
            with self.assertRaises(exception):
                self._setup_ce(req,
                               collaborators=collaborators,
                               **extra_attributes)
                cnx.commit()
            cnx.rollback()

    def _clone_ce(self, req, clone_rtype='clone_of',
                  clone_name=u'My cloned entity'):
        entity = req.entity_from_eid(self.ce_eid)
        clone = req.create_entity(entity.cw_etype, name=clone_name,
                                  description=entity.description,
                                  **{clone_rtype: entity})
        return clone.eid

    def clone_ce(self, user_eid, clone_rtype, clone_name=u'My cloned entity'):
        with self.do_as_user(user_eid) as req:
            clone_eid = self._clone_ce(req, clone_rtype, clone_name=clone_name)
        return clone_eid

    def assert_clone_ce_raises(self, exception, user_eid, clone_rtype,
                               clone_name=u'My cloned entity'):
        login = self.request().entity_from_eid(user_eid).login
        with self.userlogin(login) as cnx:
            with self.assertRaises(exception):
                self._clone_ce(cnx.request(), clone_rtype, clone_name=clone_name)
                cnx.commit()
            cnx.rollback()

    def check_clone(self, clone_eid, creator_eid, clone_rtype='clone_of',
                    clone_name=u'My cloned entity',
                    child_entities=False, other_entities=False):
        test_ent = self.request().entity_from_eid(self.ce_eid)
        cloned_ent = self.request().entity_from_eid(clone_eid)
        self.assertEqual(cloned_ent.owned_by[0].eid, creator_eid)
        for entity in cloned_ent.reverse_in_collaborativeentity:
            self.assertEqual(entity.created_by[0].eid, creator_eid)
            self.assertEqual(entity.owned_by[0].eid, creator_eid)
        self.assertEqual(cloned_ent.owned_by[0].eid, creator_eid)
        self.assertTrue(test_ent.cw_adapt_to('ICollaborative').clones)
        self.assertEqual(test_ent.name, u'My test entity')
        self.assertEqual(cloned_ent.name, clone_name)
        self.assertEqual(getattr(cloned_ent, clone_rtype)[0].eid, test_ent.eid)
        self.assertNotEqual(cloned_ent.eid, test_ent.eid)
        self.assertEqual(cloned_ent.description, test_ent.description)
        wf_clone = cloned_ent.cw_adapt_to('IWorkflowable')
        self.assertEqual(wf_clone.state, 'wfs_mutable')

        getattr_entities = lambda entities, name: frozenset(getattr(_e, name)
                                                            for _e in entities)
        if child_entities:
            self.assertEqual(len(test_ent.child_entities),
                             len(cloned_ent.child_entities))
            self.assertFalse(
                getattr_entities(test_ent.child_entities, 'eid') &
                getattr_entities(cloned_ent.child_entities, 'eid')
            )
            self.assertEqual(
                getattr_entities(test_ent.child_entities, 'name'),
                getattr_entities(cloned_ent.child_entities, 'name')
            )
            self.assertEqual(
                getattr_entities(test_ent.child_entities, 'description'),
                getattr_entities(cloned_ent.child_entities, 'description')
            )

            test_sub_entities = [ent
                                 for ent in test_ent.child_entities
                                 for scent in ent.sub_children]
            cloned_sub_entities = [ent
                                   for ent in cloned_ent.child_entities
                                   for scent in ent.sub_children]
            self.assertFalse(
                getattr_entities(test_sub_entities, 'eid') &
                getattr_entities(cloned_sub_entities, 'eid')
            )
            self.assertEqual(
                getattr_entities(test_sub_entities, 'name'),
                getattr_entities(cloned_sub_entities, 'name')
            )
            self.assertEqual(
                getattr_entities(test_sub_entities, 'description'),
                getattr_entities(cloned_sub_entities, 'description')
            )
        if other_entities:
            self.assertEqual(len(test_ent.other_entities),
                             len(cloned_ent.other_entities))
            # links to the same entities for non-composite
            self.assertEqual(getattr_entities(test_ent.other_entities, 'eid'),
                             getattr_entities(cloned_ent.other_entities, 'eid'))
            if child_entities:

                test_other_entities = [scent
                                       for ent in test_ent.child_entities
                                       for scent in ent.others]
                cloned_other_entities = [scent
                                         for ent in cloned_ent.child_entities
                                         for scent in ent.others]
                # links to the same entities for non-composite
                self.assertEqual(
                    getattr_entities(test_other_entities, 'eid'),
                    getattr_entities(cloned_other_entities, 'eid')
                )
                self.assertEqual(
                    getattr_entities(test_other_entities, 'name'),
                    getattr_entities(cloned_other_entities, 'name')
                )
                self.assertEqual(
                    getattr_entities(test_other_entities, 'other_description'),
                    getattr_entities(cloned_other_entities, 'other_description')
                )

    def setup_ce_owners(self, owners):
        """Setup owners on the collaborative entity. """

        req = self.request()
        entity = req.entity_from_eid(self.ce_eid)
        entity.cw_set(owned_by=owners)
        self.commit()




# -*- coding: utf-8 -*-
# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.


from cubicweb import ValidationError, Unauthorized, NoSelectableObject
from cubicweb.schema import ERQLExpression, RRQLExpression
from cubicweb.devtools.testlib import CubicWebTC

from cubes.container.testutils import ContainerMixinTC

from cubes.collaboration.views import CloneAction
from cubes.collaboration.testutils import CollaborationMixin


class CollaborationTC(ContainerMixinTC, CollaborationMixin, CubicWebTC):
    child_entities = True
    other_entities = True
    clone_rtype = 'clone_of_collaborative_entity'

    def setup_database(self):
        self.prepare_entities('CollaborativeEntity',
                              subetypes=('ChildEntity',),
                              otheretypes=('OtherEntity',))

    def make_clonable(self, entity):
        wf_entity = entity.cw_adapt_to('IWorkflowable')
        if wf_entity.state == 'wfs_immutable':
            return
        wf_entity.fire_transition('wft_make_immutable')
        self.commit()
        entity.cw_clear_all_caches()
        assert wf_entity.state == 'wfs_immutable', wf_entity.state


class CollaborationSchemaTC(ContainerMixinTC, CubicWebTC):

    def test_collaborative_static_structure(self):
        from config import COLLAB_CONTAINER
        container_cfg = COLLAB_CONTAINER.container_config
        rtypes, etypes = container_cfg.structure(self.schema)
        self.set_description('container static structure: rtypes')
        yield self.assertCountEqual, rtypes, ['child_entities', 'sub_children']
        self.set_description('container static structure: etypes')
        struct_etypes = ('ChildEntity', 'SubChildEntity')
        yield self.assertCountEqual, etypes, struct_etypes
        strict_etypes = container_cfg.structure(self.schema, strict=True)[1]
        self.set_description('container static structure: strict etypes')
        yield self.assertCountEqual, strict_etypes, struct_etypes

    def test_container_permissions_overwrite(self):
        etype = self.schema['CollaborativeEntity']
        rqlexprs = 'U can_read X, X name != "secret"'
        self.assertEqual(etype.permissions['read'],
                         ('managers', ERQLExpression(rqlexprs)))

    def test_etype_in_container_permissions_overwrite(self):
        self.skipTest('permissions overridding of entity type not strictly in '
                      'container not implemented')
        etype = self.schema['ChildEntity']
        update_rqlexpr = ('U can_write C, X in_collaborativeentity C, '
                          'NOT EXISTS(C in_state ST, ST name "wfs_immutable"), '
                          'X name != "don\'t touch"')
        self.assertEqual(etype.permissions['update'],
                         ('managers', ERQLExpression(update_rqlexpr)))
        read_rqlexpr = ('EXISTS(U can_read C, X in_collaborativeentity C) OR '
                        'NOT EXISTS(X in_collaborativeentity C, X owned_by U)')
        self.assertEqual(etype.permissions['read'],
                         ('managers', ERQLExpression(read_rqlexpr)))

    def test_near_container_rdef_permissions_overwrite(self):
        rdef = self.schema['child_entities'].rdefs['CollaborativeEntity', 'ChildEntity']
        add_rqlexpr = ('U can_write S, '
                       'NOT EXISTS(S in_state ST, ST name "wfs_immutable"), '
                       'S name != "I don\'t like kids"')
        self.set_description('near container rdef: rql expr')
        yield (self.assertEqual, rdef.permissions['add'],
               ('managers', RRQLExpression(add_rqlexpr)))
        # Also checks that 'users' has been removed from groups.
        self.set_description('near container rdef: groups')
        del_rqlexpr = ('U can_write S, '
                       'NOT EXISTS(S in_state ST, ST name "wfs_immutable"), '
                       'O name != "keep me"')
        yield (self.assertEqual, rdef.permissions['delete'],
               ('managers', RRQLExpression(del_rqlexpr)))

    def test_inner_rdef_permissions_overwrite(self):
        self.skipTest('permissions overridding of rdef not strictly in '
                      'container not implemented')
        rdef = self.schema['sub_children'].rdefs['ChildEntity', 'SubChildEntity']
        del_rqlexpr = (u'(NOT EXISTS(S in_collaborativeentity C1)) OR '
                       u'EXISTS(S in_collaborativeentity C, U can_write C, '
                       u'       NOT EXISTS(C in_state ST, ST name "wfs_immutable")), '
                       u'S name != "keep me"')
        self.set_description('in container rdef: rql expr')
        yield (self.assertEqual, rdef.permissions['delete'],
               ('managers', RRQLExpression(del_rqlexpr)))
        # Also check that 'users' has been removed from groups.
        self.set_description('in container rdef: groups')
        add_rqlexpr = ('(NOT EXISTS(S in_collaborativeentity C1)) OR '
                       'EXISTS(S in_collaborativeentity C, U can_write C, '
                       '       C in_state ST, NOT ST name "wfs_immutable")')
        yield (self.assertEqual, rdef.permissions['add'],
               ('managers', RRQLExpression(add_rqlexpr)))


class CollaborationWorkflowSetupTC(ContainerMixinTC, CubicWebTC):
    """Tests for collaboration workflow setup"""

    def get_etype_workflow(self, req=None):
        req = req or self.request()
        rset = self.request().execute(
            'Any ET WHERE ET is CWEType, ET name "CollaborativeEntity"')
        etype = rset.get_entity(0, 0)
        assert etype.reverse_workflow_of
        return etype, etype.reverse_workflow_of[0]

    def test_workflow_exists(self):
        etype, wf = self.get_etype_workflow(self.request())
        self.assertEqual(wf.name,
                         'collaboration workflow for CollaborativeEntity')
        self.assertEqual(etype.default_workflow[0].eid, wf.eid)

    def assertTransitionsEqual(self, wfadapter, transitions):
        self.assertEqual(
            set(transitions),
            set([t.name for t in wfadapter.possible_transitions()]))

    def test_workflow_setup(self):
        self.create_user(self.request(), u'bob')
        etype, wf = self.get_etype_workflow(self.request())
        with self.login('bob'):
            req = self.request()
            entity = req.create_entity('CollaborativeEntity', name=u'zorg',
                                       description=u'gni')
            self.commit()
            wfa = entity.cw_adapt_to('IWorkflowable')
            self.assertIsNotNone(wfa)
            self.assertEqual(wfa.current_workflow.eid, wf.eid)
            self.assertEqual(wfa.state, 'wfs_mutable')
            self.assertTransitionsEqual(wfa, ['wft_make_immutable'])
            wfa.fire_transition('wft_make_immutable')
            self.commit()
            entity.cw_clear_all_caches()
            self.assertEqual(wfa.state, 'wfs_immutable')
            self.assertTransitionsEqual(wfa, ['wft_make_mutable'])
            clone = req.create_entity('CollaborativeEntity', name=u'glub',
                                      description=u'gna',
                                      clone_of_collaborative_entity=entity)
            self.commit()
            entity.cw_clear_all_caches()
            # XXX Damned caches...
            # (see _check method of RQLExpression in cubicweb.schema)
            req.local_perm_cache.clear()
            # Not possible to make the entity mutable if it has clones.
            self.assertTransitionsEqual(wfa, [])
            with self.assertRaises(ValidationError):
                wfa.fire_transition('wft_make_mutable')
            self.rollback()


class ICollaborativeTC(CollaborationTC):

    def test_set_collaborators(self):
        req = self.request()
        entity = req.entity_from_eid(self.ce_eid)
        entity.cw_adapt_to('ICollaborative').set_collaborators(self.users())
        self.assertCountEqual(
            (user.eid for user in entity.reverse_collaborates_on),
            self.users())


class ICollaborationSecurityTC(ContainerMixinTC, CubicWebTC):

    def setup_database(self):
        self.author_eid = self.create_user(
            self.request(), 'author', groups=('users',)).eid
        with self.login('author'):
            self.entity_eid = self.request().create_entity(
                'CollaborativeEntity', name=u'My test entity',
                description=u'My description').eid

    def test_security_adapter_methods(self):
        admin_req = self.request()
        entity = admin_req.entity_from_eid(self.entity_eid)
        adapted = entity.cw_adapt_to('ICollaborationSecurity')
        # revoke read
        self.assertEqual(entity.reverse_can_read[0].eid, self.author_eid)
        adapted.revoke_permission(self.author_eid, 'read')
        entity.cw_clear_all_caches()
        self.assertEqual(len(entity.reverse_can_read), 0)
        # revoke write
        self.assertEqual(entity.reverse_can_write[0].eid, self.author_eid)
        adapted.revoke_permission(self.author_eid, 'write')
        entity.cw_clear_all_caches()
        self.assertEqual(len(entity.reverse_can_write), 0)
        # grant read
        self.assertEqual(len(entity.reverse_can_read), 0)
        adapted.grant_permission(self.author_eid, 'read')
        entity.cw_clear_all_caches()
        self.assertEqual(entity.reverse_can_read[0].eid, self.author_eid)
        # grant write; test write implies read (hence, revoke read first)
        adapted.revoke_permission(self.author_eid, 'read')
        entity.cw_clear_all_caches()
        self.assertEqual(len(entity.reverse_can_read), 0)
        adapted.grant_permission(self.author_eid, 'write')
        entity.cw_clear_all_caches()
        self.assertEqual(entity.reverse_can_write[0].eid, self.author_eid)
        self.assertEqual(entity.reverse_can_read[0].eid, self.author_eid)


class CollaborationIBreadCrumbsTC(CollaborationTC):

    def test_collaborative_breadcrumbs_for_collaborative_entities(self):
        req = self.request()
        entity = req.entity_from_eid(self.ce_eid)
        ibread_entity = entity.cw_adapt_to('IBreadCrumbs')
        self.assertEqual([entity], ibread_entity.breadcrumbs())
        one_child = entity.child_entities[0]
        ibread_child = one_child.cw_adapt_to('IBreadCrumbs')
        self.assertEqual([one_child], ibread_child.breadcrumbs())


class CollaborationActionsTC(CollaborationTC):

    def test_clone(self):
        with self.login('author'):
            req = self.request()
            entity = req.entity_from_eid(self.ce_eid)
            self.set_description('clone action not available if not public')
            yield self.check_no_clone_action, req, entity
            self.make_clonable(entity)
            self.set_description('clone action available if public')
            yield self.check_clone_action, req, entity
            self.set_description('clone edit form')
            assert entity.cw_adapt_to('Container.clone')
            yield self.check_clone_edit_form, req, entity

    def check_no_clone_action(self, req, entity):
        with self.assertRaises(NoSelectableObject):
            self.vreg['actions'].select('copy', req, rset=entity.as_rset())

    def check_clone_action(self, req, entity):
        action = self.vreg['actions'].select('copy', req,
                                             rset=entity.as_rset())
        self.assertIsInstance(action, CloneAction)

    def check_clone_edit_form(self, req, entity):
        clone_name = entity.name + ' (my clone)'
        req.form = {
            '__maineid' : 'X',
            'eid': 'X',
            '__cloned_eid:X': self.ce_eid,
            '__type:X': 'CollaborativeEntity',
            '_cw_entity_fields:X': ('name-subject,'
                                    'description-subject,'
                                    'clone_of_collaborative_entity-subject'),
            'name-subject:X': clone_name,
            'description-subject:X': entity.description,
            'clone_of_collaborative_entity-subject:X': self.ce_eid,
        }
        self.expect_redirect_handle_request(req, 'edit')
        # Ensure the clone now exists.
        clone = req.find('CollaborativeEntity', name=clone_name,
                         clone_of_collaborative_entity=entity.eid).one()
        user_eid = self.users('author')[0]
        self.check_clone(clone.eid, user_eid,
                         clone_rtype = self.clone_rtype,
                         child_entities=self.child_entities,
                         other_entities=self.other_entities,
                         clone_name=clone_name)


class CollaborationPolicyTC(CollaborationTC):

    def test_author_can_clone_if_container_clonable(self):
        # XXX To put into container
        entity = self.request().entity_from_eid(self.ce_eid)
        self.make_clonable(entity)
        clone_eid = self.clone_ce('author', clone_rtype=self.clone_rtype)
        self.check_clone(clone_eid, self.users('author')[0],
                         clone_rtype=self.clone_rtype,
                         child_entities=self.child_entities,
                         other_entities=self.other_entities)

    def test_collaborators_can_clone_if_container_clonable(self):
        entity = self.request().entity_from_eid(self.ce_eid)
        self.make_clonable(entity)
        self.setup_ce(collaborators=self.users())
        for i, collaborator_eid in enumerate(self.users()):
            clone_eid = self.clone_ce(collaborator_eid, clone_rtype=self.clone_rtype,
                                      clone_name=u'My cloned entity_%s' % i)
            self.check_clone(clone_eid,
                             collaborator_eid,
                             clone_rtype=self.clone_rtype,
                             clone_name=u'My cloned entity_%s' % i,
                             child_entities=self.child_entities,
                             other_entities=self.other_entities)

    def test_cannot_clone_unless_collaborator_or_author(self):
        entity = self.request().entity_from_eid(self.ce_eid)
        self.make_clonable(entity)
        self.setup_ce(collaborators=self.users(u'0'))
        other_user_eids = self.users(u'[^0]')
        for i, user_eid in enumerate(other_user_eids):
            self.assert_clone_ce_raises(Unauthorized, user_eid, clone_rtype=self.clone_rtype,
                                        clone_name=u'My cloned entity_%s' % i)

    def test_cannot_clone_if_in_wrong_state(self):
        entity = self.request().entity_from_eid(self.ce_eid)
        assert entity.cw_adapt_to('IWorkflowable').state == 'wfs_mutable'
        self.assert_clone_ce_raises(ValidationError, self.users('author')[0], clone_rtype=self.clone_rtype)

    def test_cannot_edit_composite_rels_if_container_cloned(self):
        req = self.request()
        entity = req.entity_from_eid(self.ce_eid)
        self.make_clonable(entity)
        clone = req.create_entity(entity.cw_etype, name=u'blah',
                                  description=entity.description,
                                  **{self.clone_rtype: entity})
        self.commit()
        with self.login('author'):
            req = self.request()
            req.create_entity('ChildEntity', name=u'New child',
                              description=u'new description',
                              reverse_child_entities=self.ce_eid)
            with self.assertRaises(Unauthorized):
                self.commit()
            self.rollback()

    def test_can_edit_composite_rels_if_container_not_cloned(self):
        with self.login('author'):
            req = self.request()
            req.create_entity('ChildEntity', name=u'New child',
                              description=u'new description',
                              reverse_child_entities=self.ce_eid)
            self.commit()

    def test_cannot_edit_sub_children_if_container_immutable(self):
        entity = self.request().entity_from_eid(self.ce_eid)
        entity.cw_adapt_to('IWorkflowable').fire_transition('wft_make_immutable')
        self.commit()
        entity.cw_clear_all_caches()
        with self.login('author'):
            req = self.request()
            one_child = req.entity_from_eid(self.ce_eid).child_entities[0]
            with self.assertRaises(Unauthorized):
                req.create_entity('SubChildEntity',
                                  name=u'New sub-child',
                                  description=u'new sub-description',
                                  reverse_sub_children=one_child)
                self.commit()
            self.rollback()

    def test_can_edit_sub_children_if_container_not_cloned(self):
        with self.login('author'):
            req = self.request()
            one_child = req.entity_from_eid(self.ce_eid).child_entities[0]
            req.create_entity('SubChildEntity',
                              name=u'New sub-child',
                              description=u'new sub-description',
                              reverse_sub_children=one_child)

    def test_can_still_edit_non_composite_rels_if_container_immutable(self):
        #self.skipTest('to be continued, not sure what this is supposed to do')
        entity = self.request().entity_from_eid(self.ce_eid)
        entity.cw_adapt_to('IWorkflowable').fire_transition('wft_make_immutable')
        self.commit()
        # XXX Is this the right behavior?
        with self.login('author'):
            req = self.request()
            new_other = req.create_entity(
                'OtherEntity',
                name=u'New child',
                other_description=u'new description',
                reverse_other_entities=self.ce_eid,
            )
            self.commit()

    def test_cannot_edit_container_if_immutable(self):
        entity = self.request().entity_from_eid(self.ce_eid)
        entity.cw_adapt_to('IWorkflowable').fire_transition('wft_make_immutable')
        self.commit()
        with self.login('author'):
            req = self.request()
            entity = req.entity_from_eid(self.ce_eid)
            yield self.check_cannot_edit_attr, entity
            yield self.check_cannot_edit_sub_entities, entity

    def check_cannot_edit_attr(self, entity):
        self.set_description('cannot edit container attributes')
        entity.cw_set(name=u'Another, new name')
        with self.assertRaises(Unauthorized):
            self.commit()
        self.rollback()

    def check_cannot_edit_sub_entities(self, entity):
        self.set_description('cannot edit entities in container')
        child = entity.child_entities[0]
        child.cw_set(description=u'New child description')
        with self.assertRaises(Unauthorized):
            self.commit()
        self.rollback()

    def test_can_edit_collaborators_even_if_container_in_clonable_state(self):
        req = self.request()
        user_1 = self.users(u'1')
        entity = req.entity_from_eid(self.ce_eid)
        entity.cw_adapt_to('IWorkflowable').fire_transition('wft_make_immutable')
        self.commit()
        # Add a collaborator
        self.setup_ce(collaborators=user_1)
        adapter = entity.cw_adapt_to('ICollaborative')
        # Check the collaborator is the only collaborator
        self.assertEqual([adapter.collaborators[0].eid], user_1)
        user_2 = self.users(u'2')
        # Add another collaborator
        self.setup_ce(collaborators=user_2)
        entity.cw_clear_all_caches()
        # Check this other collaborator is now the only collaborator
        self.assertEqual([adapter.collaborators[0].eid], user_2)
        entity.cw_clear_all_caches()
        # Pick up two collaborators
        users_1_2 = self.users(u'[12]')
        # Check that we really have two collaborators
        self.assertEqual([user for user in users_1_2],
                         [user_1[0], user_2[0]])
        # Add these two collaborators
        self.setup_ce(collaborators=users_1_2)
        # Check that these two collaborators are now the only collaborators
        self.assertEqual(frozenset(col.eid for col in adapter.collaborators),
                         frozenset(users_1_2))

    def test_collaborates_on_perms(self):
        collaborator_eid = self.create_user(self.request(), 'coll').eid
        with self.login('author'):
            req = self.request()
            entity = req.entity_from_eid(self.ce_eid)
            entity.cw_adapt_to('IWorkflowable').fire_transition('wft_make_immutable')
            self.commit()
            entity.cw_set(reverse_collaborates_on=collaborator_eid)
            self.commit()
        with self.login('coll'):
            req = self.request()
            clone_eid = req.create_entity(
                'CollaborativeEntity', name=u'my clone',
                description=u'same here',
                **{self.clone_rtype: self.ce_eid}).eid
            self.commit()
        with self.login('author'):
            req = self.request()
            entity = req.entity_from_eid(self.ce_eid)
            with self.assertRaises(Unauthorized):
                # Cannot remove collaborator if she has a clone.
                entity.cw_set(reverse_collaborates_on=None)
                self.commit()
            self.rollback()
        with self.login('coll'):
            req = self.request()
            clone = req.entity_from_eid(clone_eid)
            clone.cw_delete()
            self.commit()
        with self.login('author'):
            req = self.request()
            entity = req.entity_from_eid(self.ce_eid)
            # No more clone, can remove collaborator.
            entity.cw_set(reverse_collaborates_on=None)
            self.commit()

    def test_cannot_read_unless_author_or_collaborator(self):
        self.setup_ce(collaborators=self.users(u'0'))
        other_user_eids = self.users(u'[^0]')
        for user_eid in other_user_eids:
            login = self.request().entity_from_eid(user_eid).login
            with self.userlogin(login) as cnx:
                req = cnx.request()
                with self.assertRaises(Unauthorized):
                    # req.entity_from_eid uses req.eid_rset that does not
                    # perform a real request and thus never raises Unauthorized
                    req.execute('Any X WHERE X eid %(eid)s',
                                {'eid': self.ce_eid})
                cnx.rollback()

    def test_author_can_edit_clone(self):
        author_eid = self.users('author')[0]
        self.make_clonable(self.request().entity_from_eid(self.ce_eid))
        clone_eid = self.clone_ce(author_eid, clone_rtype=self.clone_rtype)
        with self.do_as_user('author') as req:
            clone = req.entity_from_eid(clone_eid)
            clone.cw_set(name=u'New clone name')

    def test_collaborators_can_edit_clone(self):
        entity = self.request().entity_from_eid(self.ce_eid)
        self.make_clonable(entity)
        collaborator_eids = self.users()
        self.setup_ce(collaborators=collaborator_eids)
        for i, collaborator_eid in enumerate(collaborator_eids):
            clone_eid = self.clone_ce(collaborator_eid, clone_rtype=self.clone_rtype)
            with self.do_as_user(collaborator_eid) as req:
                clone = req.entity_from_eid(clone_eid)
                clone.cw_set(name=u'New clone name %s' % i)

    def test_collaborators_cannot_edit_container(self):
        collaborator_eids = self.users()
        self.setup_ce(collaborators=collaborator_eids)
        for i, collaborator_eid in enumerate(collaborator_eids):
            login = self.request().entity_from_eid(collaborator_eid).login
            with self.userlogin(login) as cnx:
                req = cnx.request()
                original_ent = req.entity_from_eid(self.ce_eid)
                with self.assertRaises(Unauthorized):
                    original_ent.cw_set(name=u'New name %s for the original' % i)
                    cnx.commit()
                cnx.rollback()

    def test_mutable_transitions_clone_workflow(self):
        req = self.request()
        user = self.create_user(req, u'toto')
        entity = req.entity_from_eid(self.ce_eid)
        entity.cw_set(owned_by=user)
        self.commit()
        with self.login('toto'):
            req = self.request()
            entity = req.entity_from_eid(self.ce_eid)
            # Not immutable, can modify.
            entity.cw_set(name=u'toto here')
            wfa = entity.cw_adapt_to('IWorkflowable')
            self.commit()
            # Make immutable, cannot modify.
            wfa.fire_transition('wft_make_immutable')
            self.commit()
            with self.assertRaises(Unauthorized):
                entity.cw_set(name=u'toto again')
                self.commit()
            self.rollback()
            # Make mutable, and modify.
            wfa.fire_transition('wft_make_mutable')
            self.commit()
            entity.cw_set(name=u'toto again')
            self.commit()
            # Make immutable and clone.
            wfa.fire_transition('wft_make_immutable')
            self.commit()
            clone = req.create_entity('CollaborativeEntity', name=u'un clone',
                                      description=u'tout pareil',
                                      clone_of_collaborative_entity=entity)
            self.commit()
            # Cannot make the original mutable.
            with self.assertRaises(ValidationError) as err:
                wfa.fire_transition('wft_make_mutable')
            self.rollback()
            self.assertIn('transition may not be fired', str(err.exception))
            # No more clone, can make mutable.
            clone.cw_delete()
            self.commit()
            wfa.fire_transition('wft_make_mutable')
            self.commit()

    def test_owners_cannot_delete_others_clones(self):
        owners = self.users(u'[0-2]')
        self.setup_ce_owners(owners)
        admin_req = self.request()
        entity = admin_req.entity_from_eid(self.ce_eid)
        self.make_clonable(entity)
        for owner_eid in owners:
            clone_eid = self.clone_ce(owner_eid, clone_rtype=self.clone_rtype,
                                      clone_name=u'%s\'s new name' % owner_eid)
            with self.login(admin_req.entity_from_eid(owner_eid).login):
                req = self.request()
                clone_etype = req.describe(clone_eid)[0]
                # get clones not owned by current owner
                rql = ('Any E WHERE X is %(cetype)s, X eid E, '
                       'EXISTS(X %(clone_rtype)s Y), '
                       'NOT X owned_by U, U eid %%(ueid)s'
                       % {'cetype': clone_etype,
                          'clone_rtype': self.clone_rtype})
                other_clones = admin_req.execute(rql, {'ueid': owner_eid}).rows
                for other_clone in (oc[0] for oc in other_clones):
                    with self.assertRaises(Unauthorized):
                        req.execute('DELETE %(cetype)s X WHERE X eid %%(ceid)s'
                                    % {'cetype': clone_etype},
                                    {'ceid': other_clone})
                        self.commit()
                    self.rollback()

    def test_former_owners_cannot_read_unless_collaborators(self):
        owners = self.users(u'[0-2]')
        self.setup_ce_owners(owners)
        admin_req = self.request()
        entity = admin_req.entity_from_eid(self.ce_eid)
        self.setup_ce(collaborators=owners[0])
        for owner_eid in owners:
            admin_req.execute('DELETE X owned_by U WHERE X eid %(eeid)s, U eid %(ueid)s',
                              {'eeid': self.ce_eid, 'ueid': owner_eid})
            self.commit()
            self.assert_setup_ce_raises(Unauthorized, as_user=owner_eid, name=u'my other_name')
            with self.userlogin(admin_req.entity_from_eid(owner_eid).login) as cnx:
                req = cnx.request()
                if owner_eid == owners[0]:
                    my_entity = req.execute('Any X WHERE X eid %(eeid)s',
                                            {'eeid': self.ce_eid}).get_entity(0, 0)
                    self.assertEqual(my_entity.cw_etype, entity.cw_etype)
                    self.assertEqual(frozenset(ch.eid for ch in my_entity.child_entities),
                                     frozenset(ch.eid for ch in  entity.child_entities))
                else:
                    with self.assertRaises(Unauthorized):
                        req.execute('Any X WHERE X eid %(eeid)s', {'eeid': self.ce_eid})

    def test_owners_can_edit_collaborators(self):
        owners = self.users(u'[0-2]')
        self.setup_ce_owners(owners)
        entity = self.request().entity_from_eid(self.ce_eid)
        adapted = entity.cw_adapt_to('ICollaborative')
        for owner_eid in owners:
            self.setup_ce(as_user=owner_eid, collaborators=())
            entity.cw_clear_all_caches()
            self.assertEqual(len(adapted.collaborators), 0)
            self.setup_ce(as_user=owner_eid, collaborators=owners[:2])
            entity.cw_clear_all_caches()
            self.assertEqual(frozenset(c.eid for c in adapted.collaborators),
                             frozenset(owners[:2]))




if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-collaboration tests of utilities"""

from unittest import TestCase

from yams.buildobjs import EntityType, RelationDefinition
from yams.reader import build_schema_from_namespace

from cubicweb.schema import ERQLExpression, RRQLExpression

# Import "extended" EntityType and RelationDefinition classes.
import cubes.collaboration


class SchemaPermissionsTC(TestCase):
    """Test cases for schema classes permissions extension utilities"""

    def setUp(self):
        class Clown(EntityType):
            __permissions__ = {
                'read':   ('managers', ERQLExpression('X title N'),
                           ERQLExpression('X sad True')),
                'add':    ('managers', 'users'),
                'update': ('managers',),
                'delete': ('managers', 'owners'),
                }

        class Joke(EntityType):
            """Funny"""

        class tricks(RelationDefinition):
            __permissions__ = {
                'read': ('managers', 'users', 'guests'),
                'add': ('managers', 'users'),
                'delete': ('managers', RRQLExpression('S title N'),
                           RRQLExpression('O funny True'))
            }
            subject = 'Clown'
            object  = 'Joke'

        self.schema = build_schema_from_namespace(locals().items())

    def test_redundant_variable(self):
        etype = self.schema['Clown']
        with self.assertRaises(Exception) as cm:
            etype.permissions_add_mandatory_rqlexpression('read', 'X color N')
        self.assertEqual(str(cm.exception),
                         'Variable `N` already used in RQL expression '
                         '`X title N`')

    def test_etype_add_mandatory_rqlexpr(self):
        etype = self.schema['Clown']
        etype.permissions_add_mandatory_rqlexpression(
            'read', 'X color V, V is "red"')
        self.assertEqual(etype.permissions['read'],
                         ('managers',
                          ERQLExpression('X color V, V is "red", X title N'),
                          ERQLExpression('X color V, V is "red", X sad True')))

    def test_etype_add_mandatory_rqlexpr_no_rqlexpr_yet(self):
        etype = self.schema['Clown']
        etype.permissions_add_mandatory_rqlexpression('add', 'X color V')
        self.assertEqual(etype.permissions['add'],
                         ('managers', 'users', ERQLExpression("X color V")))

    def test_etype_set_permissions_groups(self):
        etype = self.schema['Clown']
        etype.set_permissions_groups('read', ('clowns', 'monkeys'))
        self.assertEqual(etype.permissions['read'],
                         ('clowns', 'monkeys', ERQLExpression('X title N'),
                          ERQLExpression('X sad True')))

    def test_rdef_add_mandatory_rqlexpr(self):
        rdef = self.schema['tricks'].rdefs['Clown', 'Joke']
        rdef.permissions_add_mandatory_rqlexpression(
            'delete', 'S name "bozo"')
        self.assertEqual(rdef.permissions['delete'],
                         ('managers',
                          RRQLExpression('S name "bozo", S title N'),
                          RRQLExpression('S name "bozo", O funny True')))

    def test_rdef_add_mandatory_rqlexpr_no_rqlexpr_yet(self):
        rdef = self.schema['tricks'].rdefs['Clown', 'Joke']
        rdef.permissions_add_mandatory_rqlexpression('add', 'S color V')
        self.assertEqual(rdef.permissions['add'][-1],
                         RRQLExpression("S color V"),)

    def test_rdef_set_permissions_groups(self):
        rdef = self.schema['tricks'].rdefs['Clown', 'Joke']
        rdef.set_permissions_groups('read', ('clowns', 'monkeys'))
        self.assertEqual(rdef.permissions['read'],
                         ('clowns', 'monkeys'))


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()
